/** I N C L U D E S **********************************************************/
#include <p18cxxx.h>
#include "delays.h"
#include "system\typedefs.h"
#include "system\usb\usb.h"

#include "io_cfg.h"                 // I/O pin mapping
#include "user\pickit.h"
#include "user\pk_isr.h"

/** V A R I A B L E S ********************************************************/

#pragma udata	USB_Buffers
char inbuffer[BUF_SIZE];            				// input to USB device buffer
char outbuffer[BUF_SIZE];            				// output to USB device buffer

#pragma udata	DownloadBuffer
unsigned char 	uc_download_buffer[DOWNLOAD_SIZE];	// Download Data Buffer

#pragma udata	UploadBuffer
unsigned char 	uc_upload_buffer[UPLOAD_SIZE];		// Upload Data Buffer

#pragma udata access MiscAccess

#pragma udata	Misc
struct {
    unsigned char	write_index;        // buffer write index
    unsigned char	read_index;			// buffer read index
    unsigned char   used_bytes;         // # bytes in buffer
} downloadbuf_mgmt;

struct {
    unsigned char	write_index;		// buffer write index
    unsigned char	read_index;			// buffer read index
    unsigned char   used_bytes;         // # bytes in buffer
} uploadbuf_mgmt; 

union {		// Status bits
	struct {
		unsigned char	StatusLow;
		unsigned char	StatusHigh;
	};
	struct{
		// StatusLow
		unsigned ExtPsConnected:1;	// bit 0
		unsigned ButtonPressed:1;
		unsigned BusyLedOn:1;
		unsigned CmdAnswared:1;
		unsigned :1;
		unsigned :1;
        unsigned :1;
		unsigned Reset:1;
		//StatusHigh
        unsigned UpLoadUndFlow:1;       // bit 0
		unsigned UpLoadEmpty:1;			
        unsigned UpLoadFull:1;
		unsigned UpLoadOvrFlow:1;
		unsigned DownLoadUndFlow:1;
		unsigned DownLoadEmpty:1;
		unsigned DownLoadFull:1;
		unsigned DownLoadOvrFlow:1;
	};
} Status;


#pragma udata	Misc2

#pragma udata	Misc3

#pragma udata	Misc4



/** P R I V A T E  P R O T O T Y P E S ***************************************/

unsigned char ReadDownloadBuffer(void);
void WriteDownloadDataBuffer(unsigned char *usbindex);
void WriteUploadBuffer(unsigned char byte2write);



/******************************************************************************
 * Function:        void PickitInit(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        
 *
 * Note:            None
 *****************************************************************************/
void PICkitInit(void)
{

    //-------------------------
    // initialize I/O (see io_cfg.h)
    //-------------------------


    // ------
    // Port C
    // ------

	TRISC = 0xFF;

//  Must be input, see datasheet
//	tris_TXD = OUTPUT_PIN;			// USART transmit data pin 

    //-------------------------
    // initialize UART
    //-------------------------

	BAUDCON = 0;

/*	// Baud rate generator		 9600 Baud, 8 bit, No parity, 1 Stop
	//								48000000 / 64 / 78 = 9600
	TXSTA 	= 0;
	SPBRG 	= 77;
*/

	// Baud rate generator		100000 Baud, 8 bit, No parity, 1 Stop
	//								48000000 / 16 / 30 = 100000
	TXSTA 	= 4;
	SPBRG 	= 29;


	// Transmitter
	TXSTAbits.TXEN = 1;					// Enable transmitter

	// Receiver
	RCSTA = 0x90;					// Enable EUSART, Enable receiver, 

	c = RCREG;						// Read rcreg to clear data

    //-------------------------
    // initialize variables
    //-------------------------


    ClearDownloadBuffer();

    ClearUploadBuffer(); 



	PIE1bits.RCIE = 1;				// Enable  UART receive  IT
	PIE1bits.TXIE = 0;				// Disable UART transmit IT

	PIR2 = 0;						// Clear pending IRQ status

	// Set up Interrupts
    INTCONbits.PEIE = 1;            // enable Peripheral Interrupt Enable
    INTCONbits.GIE = 1;             // enable global interrupts

}


/******************************************************************************
 * Function:        void ProcessIO(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        
 *
 * Note:            None
 *****************************************************************************/
void ProcessIO(void)
{
	unsigned char usb_idx = 0;			// index of data in USB buffer used by WriteDownloadBuffer
	unsigned char usb_data;
											// Get next command

	if (HIDRxReport(inbuffer, 64) > 0)		// USB receive buffer has data
   	{
			usb_data = inbuffer[usb_idx];
			switch(usb_data)				// Look for commands			
			{
                case RESET: 				// Reset	 			'r'
											// Format:		0x72
											// response:	-
                    Reset();
				break;

                case GET_RESULT:			// Get result 			'g'
											// Format:		0x67
											// response:	command answare form uploadbuffer or status
                    if ( (uploadbuf_mgmt.used_bytes>0) )
					{
						ReadUploadDataBufferNoLength();
					}
					else
					{
	                    SendStatusUSB();
					}
				break;

				default:
					if ( (usb_data>'@') && (usb_data<='Z') )
					{
						WriteDownloadDataBuffer(&usb_idx);

					}					
				break;

			} // end switch

	} // end if (usb_data)
    
} // end void ProcessIO(void)




/******************************************************************************
 * Function:        void ClearUploadBuffer(void)
 *
 * Overview:        Clears the Upload Buffer
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          
 *
 * Side Effects:    
 *
 * Note:            None
 *****************************************************************************/
void ClearUploadBuffer(void)
{
    uploadbuf_mgmt.write_index = 0;     // init buffer to empty
    uploadbuf_mgmt.read_index = 0;
    uploadbuf_mgmt.used_bytes = 0; 
	Status.UpLoadOvrFlow = 0;
	Status.UpLoadEmpty = 1;
	Status.UpLoadFull = 0;
	Status.UpLoadUndFlow = 0;
}

/******************************************************************************
 * Function:        void ClearDownloadBuffer(void)
 *
 * Overview:        Clears the Download Buffer
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          
 *
 * Side Effects:    
 *
 * Note:            None
 *****************************************************************************/
void ClearDownloadBuffer(void)
{
    downloadbuf_mgmt.write_index = 0;   // init buffer to empty
    downloadbuf_mgmt.read_index = 0;
    downloadbuf_mgmt.used_bytes = 0;
	Status.DownLoadOvrFlow = 0;
	Status.DownLoadEmpty = 1;
	Status.DownLoadFull = 0;
	Status.DownLoadUndFlow = 0;
}


/******************************************************************************
 * Function:        void ReadUploadDataBufferNoLength(void)
 *
 * Overview:        Sends data from upload data buffer over USB,
 *                  but does not add a length byte.
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          Transmits HID Tx report with data only.
 *
 * Side Effects:    None
 *
 * Note:            None
 *****************************************************************************/
void ReadUploadDataBufferNoLength(void)
{
    unsigned char i, length;

    length = uploadbuf_mgmt.used_bytes;

    if (length==0)	Status.UpLoadUndFlow = 1;

    if (length > (BUF_SIZE))        		// limited to # bytes in USB report
    {
        length = (BUF_SIZE);
    }

    for (i = 0; i < length; i++)
    {
	    outbuffer[i] = uc_upload_buffer[uploadbuf_mgmt.read_index++];
        if (uploadbuf_mgmt.read_index >= UPLOAD_SIZE)  // manage buffer wrap.
        {
            uploadbuf_mgmt.read_index = 0;
        }

    }

    uploadbuf_mgmt.used_bytes -= length;    // read out this many bytes.

	Status.UpLoadFull = 0;
	if (uploadbuf_mgmt.used_bytes==0)
	{
		Status.DownLoadEmpty = 1;
	}

    // transmit data
    USBHIDTxBlocking();
} // end void ReadUploadDataBufferNoLength(void)



/******************************************************************************
 * Function:        void WriteDownloadDataBuffer(unsigned char *usbindex)
 *
 * Overview:        Writes a given # of bytes into the data download buffer.
 *
 * PreCondition:    None
 *
 * Input:           *usbindex - index to length of data in USB buffer
 *
 * Output:          uc_download_buffer[] - updated with new data
 *                  downloadbuf_mgmt.write_index - incremented by length of data stored.
 *                  downloadbuf_mgmt.used_bytes - incremented by length of data stored. 
 *                  Status.DownloadOvrFlow - set if data length > remaining buffer
 *
 * Side Effects:    None
 *
 * Note:            None
 *****************************************************************************/
void WriteDownloadDataBuffer(unsigned char *usbindex)
{
    unsigned char length=0;

	while ( (inbuffer[(*usbindex)]!='\n')&& ((*usbindex)<64) && (downloadbuf_mgmt.used_bytes<DOWNLOAD_SIZE ))
    {
       	uc_download_buffer[downloadbuf_mgmt.write_index++] = inbuffer[(*usbindex)++];
       	if (downloadbuf_mgmt.write_index >= DOWNLOAD_SIZE) // handle index wrap
       	{
           	downloadbuf_mgmt.write_index = 0;
       	}
       	length++;  							// used another byte.
    }

	PIE1bits.TXIE = 0;						// Disable UART transmit IT

   	downloadbuf_mgmt.used_bytes+= length;	// written in this many byte.

	Status.DownLoadEmpty = 0;

	if (downloadbuf_mgmt.used_bytes==DOWNLOAD_SIZE)
	{
		Status.DownLoadFull = 1;
	}

	if ((inbuffer[(*usbindex)]!='\n') && (downloadbuf_mgmt.used_bytes == DOWNLOAD_SIZE))
	{
		Status.DownLoadOvrFlow = 1;
	}

	PIE1bits.TXIE = 1;						// Enable UART transmit IT

} // end void WriteDownloadDataBuffer(unsigned char *usbindex)


/******************************************************************************
 * Function:        void WriteUploadBuffer(unsigned char byte2write)
 *
 * Overview:        Attempts to write a byte to the upload buffer.
 *                  If full, sets error Status.UpLoadFull
 * PreCondition:    None
 *
 * Input:           byte2write - byte to be written
 *
 * Output:          uc_upload_buffer - byte written to end of buffer.
 *
 * Side Effects:    Advances download buffer write pointer, if err (Status.StatusHigh & 0x0f)!= 0
 *
 * Note:            Called from isr
 *****************************************************************************/
void WriteUploadBuffer(unsigned char byte2write)
{

    if (uploadbuf_mgmt.used_bytes == UPLOAD_SIZE)     // not enough room for data
    {
        Status.UpLoadOvrFlow = 1;
    }
	else
	{
    	uc_upload_buffer[uploadbuf_mgmt.write_index++] = byte2write;
    	if (uploadbuf_mgmt.write_index >= UPLOAD_SIZE) // handle index wrap
    	{
        	uploadbuf_mgmt.write_index = 0;
    	}
    	uploadbuf_mgmt.used_bytes++;  // used another byte.

		Status.DownLoadEmpty = 0;
		if (uploadbuf_mgmt.used_bytes==UPLOAD_SIZE)
		{
			Status.UpLoadFull = 1;
		}
	}

	if (byte2write == '\n') Status.CmdAnswared = 1;
}

/******************************************************************************
 * Function:        unsigned char ReadDownloadBuffer(void)
 *
 * Overview:        Attempts to pull a byte from the Download Buffer.
 *                  If empty, sets error Status.DownloadEmpty
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          Returns byte from top of buffer.
 *
 * Side Effects:    Advances download buffer read pointer, if err Status.StatusHigh != 0
 *
 * Note:            Called from isr
 *****************************************************************************/
unsigned char ReadDownloadBuffer(void)
{
    unsigned char readbyte;

    if (downloadbuf_mgmt.used_bytes == 0)
    {
        Status.DownLoadUndFlow = 1;
        return 0;
    } 

    readbyte = uc_download_buffer[downloadbuf_mgmt.read_index++];

    downloadbuf_mgmt.used_bytes--;        				// just removed a byte.

    if (downloadbuf_mgmt.read_index >= DOWNLOAD_SIZE)   // circular buffer - handle wrap.
        downloadbuf_mgmt.read_index = 0; 

	Status.DownLoadFull = 0;
	if (downloadbuf_mgmt.used_bytes==0)
	{
		Status.DownLoadEmpty = 1;
	}

    return  readbyte;  

}


